import "../polyfill/asyncIterator.js";
import { asyncIterator as m } from "../ponyfill/asyncIterator.js";
import "../polyfill/fromAnyIterable.js";
import { fromAnyIterable as a } from "../ponyfill/fromAnyIterable.js";
export {
  m as asyncIterator,
  a as fromAnyIterable
};
