/**
 * Create a new readable stream from an async iterable or a sync iterable
 * @param iterable
 * @returns a readable stream
 */
export declare function fromAnyIterable<R>(iterable: Iterable<R> | AsyncIterable<R>): ReadableStream<R>;
